#!/usr/bin/env python3

import nbformat
import unittest
import sys
import stanza
import spacy

from tmc import points
from tmc.applt import write_result, prepare_cell, Colors

# Define path to Notebook
nb_path = 'stanza_basics.ipynb'

# Read the notebook
nb = nbformat.read(nb_path, as_version=4)

# Collect indices for notebook cells that contain exercises. Map these indices
# to a dictionary keyed by test names.
exercises = {cell['metadata']['test']: ix for ix, cell in enumerate(nb.cells)
             if 'test' in cell['metadata']}


class ImportStanza(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare the cell by clearing outputs
        prepare_cell(nb.cells[exercises['import_stanza']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['import_stanza']]

    @points('p9.1.import_module')
    def test_import_module(self):

        # Perform test
        try:

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert 'stanza' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The Stanza library was imported '
                                     f'successfully! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The Stanza library was not imported successfully. Did you refer to the ' \
                     'library using the correct name?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class LoadModel(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare the cell by clearing outputs
        prepare_cell(nb.cells[exercises['load_model']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['load_model']]

    @points('p9.2.variable_exists')
    def test_variable_exists(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert 'nlp' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "nlp" was defined '
                                     f'successfully! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "nlp" has not been defined. Did you assign the model under ' \
                     'the correct variable?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.2.variable_type', 'p9.2.variable_type_x2', 'p9.2.variable_type_x3')
    def test_variable_type(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert isinstance(locals()['nlp'], stanza.pipeline.core.Pipeline)

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "nlp" contains a Stanza '
                                     f'Pipeline object! 3 points.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "nlp" does not contain a Stanza Pipeline object! Did you ' \
                     'load the model correctly?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.2.variable_value')
    def test_variable_value(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert locals()['nlp'].lang == 'fi'

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "nlp" contains a Pipeline '
                                     f'for Finnish! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "nlp" does not contain a Pipeline for Finnish. Did you load ' \
                     'the correct model?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        except AttributeError:

            # Define error message
            errmsg = 'The variable "nlp" does not have an attribute named "lang". Did you ' \
                     'store the Pipeline under the variable "nlp"?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class ProcessText(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare cell
        prepare_cell(nb.cells[exercises['process_text']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['process_text']]

    @points('p9.3.variable_exists')
    def test_variable_exists(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check if variable exists
            assert 'my_doc' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "my_doc" was '
                                     f'defined successfully! 1 point.'
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_doc" has not been defined. Remember to ' \
                     'store the Document object under the correct variable!'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.3.variable_type', 'p9.3.variable_type_x2')
    def test_variable_type(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert isinstance(locals()['my_doc'], stanza.models.common.doc.Document)

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "my_doc" contains a Stanza '
                                     f'Document object! 2 points.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_doc" does not contain a Stanza Document object! Did you ' \
                     'provide the input data correctly to the language model?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class GetSentence(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare cell
        prepare_cell(nb.cells[exercises['get_sent']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['get_sent']]

    @points('p9.4.variable_exists')
    def test_variable_exists(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check if variable exists
            assert 'my_sent' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "my_sent" was '
                                     f'defined successfully! 1 point.'
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_sent" has not been defined. Remember to ' \
                     'store the Sentence object under the correct variable!'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.4.variable_value')
    def test_variable_value(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert locals()['my_sent'].text == 'Tässä on yksi esimerkki.'

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "my_sent" contains the '
                                     f'expected objects! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_sent" does not contain the expected objects. Did you ' \
                     'retrieve the correct sentence?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        except AttributeError:

            # Define error message
            errmsg = 'The variable "my_sent" does not have an attribute named "text". Did you ' \
                     'store the Sentence object under this variable?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.4.variable_type')
    def test_variable_type(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert isinstance(locals()['my_sent'], stanza.models.common.doc.Sentence)

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "my_sent" contains a Stanza '
                                     f'Sentence object! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_sent" does not contain a Stanza Sentence object! Did you ' \
                     'provide the input data correctly to the language model?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class ConvertDoc(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare cell
        prepare_cell(nb.cells[exercises['to_dict']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['to_dict']]

    @points('p9.5.variable_exists')
    def test_variable_exists(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)
            exec(nb.cells[exercises['get_sent']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check if variable exists
            assert 'my_list' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "my_list" was '
                                     f'defined successfully! 1 point.'
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_list" has not been defined. Remember to ' \
                     'store the resulting dictionaries under the correct variable!'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.5.variable_value')
    def test_variable_value(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)
            exec(nb.cells[exercises['get_sent']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert locals()['my_list'][0]['text'] == 'Tässä'
            assert locals()['my_list'][-1]['text'] == '.'

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "my_list" contains the '
                                     f'expected objects! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_list" does not contain the expected objects. Did you ' \
                     'remember to convert the annotations into dictionaries?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class GetLemmas(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Prepare cell
        prepare_cell(nb.cells[exercises['get_lemmas']])

    def setUp(self):
        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['get_lemmas']]

    def test_variable_exists(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)
            exec(nb.cells[exercises['get_sent']].source)
            exec(nb.cells[exercises['to_dict']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check if variable exists
            assert 'lemmas' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "lemmas" exists!'
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "lemmas" has not been defined. Did you store the result under ' \
                     'the correct variable?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    def test_variable_type(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)
            exec(nb.cells[exercises['get_sent']].source)
            exec(nb.cells[exercises['to_dict']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert isinstance(locals()['lemmas'], list)

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "lemmas" contains a list! '
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "lemmas" does not contain a list! Did you modify the ' \
                     'example code?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.6.variable_value', 'p9.6.variable_value_x2', 'p9.6.variable_value_x3',
            'p9.6.variable_value_x4')
    def test_variable_value(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['process_text']].source)
            exec(nb.cells[exercises['get_sent']].source)
            exec(nb.cells[exercises['to_dict']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            test_data = ['tämä', 'olla', 'yksi', 'esimerkki', '.']

            # Check condition
            assert locals()['lemmas'] == test_data

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "lemmas" contains the '
                                     f'expected values! 5 points.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "my_list" does not contain the expected objects. Did you ' \
                     'retrieve the lemmas from the dictionaries in the list named "my_list"?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class ImportSpaCyStanza(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare the cell by clearing outputs
        prepare_cell(nb.cells[exercises['import_spacy_stanza']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['import_spacy_stanza']]

    @points('p9.7.import_module')
    def test_import_module(self):

        # Perform test
        try:

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert 'spacy_stanza' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The spacy-stanza library was imported '
                                     f'successfully! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The spacy-stanza library was not imported successfully. Did you refer to the ' \
                     'library using the correct name?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class LoadModelSpacy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare the cell by clearing outputs
        prepare_cell(nb.cells[exercises['load_smodel']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['load_smodel']]

    @points('p9.8.variable_exists')
    def test_variable_exists(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_spacy_stanza']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check if variable exists
            assert 'spacy_fi' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "spacy_fi" was '
                                     f'defined successfully! 1 point.'
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "spacy_fi" has not been defined. Did you store the language ' \
                     'model under the correct variable?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.8.variable_type', 'p9.8.variable_type_x2')
    def test_variable_type(self):

        # Perform test
        try:

            # Run the previous exercise cell for preliminaries
            exec(nb.cells[exercises['import_spacy_stanza']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert isinstance(locals()['spacy_fi'], spacy.lang.fi.Finnish)

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The Stanza language model for Finnish was '
                                     f'loaded successfully into spaCy! 2 points.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "spacy_fi" does not contain a spaCy Language object! Did you ' \
                     'load the Stanza language model correctly using spacy-stanza?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except AttributeError:

            # Define error message
            errmsg = 'The variable "spacy_fi" does not contain a spaCy Language object! Did you ' \
                     'load the Stanza language model correctly using spacy-stanza?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Fail silently. The previous test will raise an error message.
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)


class ProcessTextSpacy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Prepare the cell by clearing outputs
        prepare_cell(nb.cells[exercises['process_text_2']])

    def setUp(self):

        # Get the cell that is to be tested
        self.cell = nb.cells[exercises['process_text_2']]

    @points('p9.9.variable_exists')
    def test_variable_exists(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['import_spacy_stanza']].source)
            exec(nb.cells[exercises['load_smodel']].source)
            exec(nb.cells[exercises['process_text']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check if variable exists
            assert 'doc_fi' in locals()

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "doc_fi" was '
                                     f'defined successfully! 1 point.'
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "doc_fi" has not been defined. Did you store the Document ' \
                     'object under the correct variable?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.9.variable_type')
    def test_variable_type(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['import_spacy_stanza']].source)
            exec(nb.cells[exercises['load_smodel']].source)
            exec(nb.cells[exercises['process_text']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert isinstance(locals()['doc_fi'], spacy.tokens.doc.Doc)

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "doc_fi" contains a spaCy '
                                     f'Doc object! 1 point.'
                                     f'{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "doc_fi" does not contain a spaCy Doc object. Did you provide ' \
                     'the input correctly to the language model?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)

    @points('p9.9.variable_value')
    def test_variable_value(self):

        # Perform test
        try:

            # Run the previous exercise cells for preliminaries
            exec(nb.cells[exercises['import_stanza']].source)
            exec(nb.cells[exercises['load_model']].source)
            exec(nb.cells[exercises['import_spacy_stanza']].source)
            exec(nb.cells[exercises['load_smodel']].source)
            exec(nb.cells[exercises['process_text']].source)

            # Run the cell within the function for local scope
            exec(self.cell.source, globals(), locals())

            # Check condition
            assert locals()['doc_fi'].text == 'Tässä on yksi esimerkki. Ja tässä on toinen.'

            # Define message for passing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.OKGREEN}The variable "doc_fi" contains the '
                                     f'expected value ! 1 point.{Colors.ENDC}'}

            # Set flag
            fail = False

        except AssertionError:

            # Define error message
            errmsg = 'The variable "doc_fi" does not contain the expected value. Did you provide ' \
                     'the input to the language model in the correct form?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        except NameError:

            # Define error message
            errmsg = 'Some variables remain undefined. Have you already ' \
                     'completed the previous exercises in this Notebook?'

            # Define a message for failing the test
            message = {'text/plain': f'{Colors.BOLD}[TMC]{Colors.ENDC} '
                                     f'{Colors.FAIL}{errmsg}{Colors.ENDC}'}

            # Set flag
            fail = True

        # Write the result to the cell
        write_result(nb=nb,
                     nb_path=nb_path,
                     cell=self.cell,
                     output_type='execute_result',
                     data=message,
                     execution_count=self.cell.execution_count)

        # Check flag
        if fail:

            # Use the fail method to fail the test
            self.fail(errmsg)
